@@ -1,8 +1,13 @@
+/* SPDX-License-Identifier: LGPL-2.1-or-later */
 /*
- * Copyright (C) 2008,2009,2012 Karel Zak <kzak@redhat.com>
+ * This file is part of libmount from util-linux project.
  *
- * This file may be redistributed under the terms of the
- * GNU Lesser General Public License.
+ * Copyright (C) 2009-2018 Karel Zak <kzak@redhat.com>
+ *
+ * libmount is free software; you can redistribute it and/or modify it
+ * under the terms of the GNU Lesser General Public License as published by
+ * the Free Software Foundation; either version 2.1 of the License, or
+ * (at your option) any later version.
  */
 
 /**
@@ -15,11 +20,7 @@
  */
 #include <ctype.h>
 
-#ifdef HAVE_LIBSELINUX
-#include <selinux/selinux.h>
-#include <selinux/context.h>
-#endif
-
+#include "strutils.h"
 #include "mountP.h"
 
 /*
@@ -33,83 +34,11 @@ struct libmnt_optloc {
 	size_t  namesz;
 };
 
-#define mnt_init_optloc(_ol)	(memset((_ol), 0, sizeof(struct libmnt_optloc)))
+#define MNT_INIT_OPTLOC	{ .begin = NULL }
 
 #define mnt_optmap_entry_novalue(e) \
 		(e && (e)->name && !strchr((e)->name, '=') && !((e)->mask & MNT_PREFIX))
 
-/*
- * Parses the first option from @optstr. The @optstr pointer is set to the beginning
- * of the next option.
- *
- * Returns -EINVAL on parse error, 1 at the end of optstr and 0 on success.
- */
-static int mnt_optstr_parse_next(char **optstr,	 char **name, size_t *namesz,
-					char **value, size_t *valsz)
-{
-	int open_quote = 0;
-	char *start = NULL, *stop = NULL, *p, *sep = NULL;
-	char *optstr0;
-
-	assert(optstr);
-	assert(*optstr);
-
-	optstr0 = *optstr;
-
-	if (name)
-		*name = NULL;
-	if (namesz)
-		*namesz = 0;
-	if (value)
-		*value = NULL;
-	if (valsz)
-		*valsz = 0;
-
-	/* trim leading commas as to not invalidate option
-	 * strings with multiple consecutive commas */
-	while (optstr0 && *optstr0 == ',')
-		optstr0++;
-
-	for (p = optstr0; p && *p; p++) {
-		if (!start)
-			start = p;		/* beginning of the option item */
-		if (*p == '"')
-			open_quote ^= 1;	/* reverse the status */
-		if (open_quote)
-			continue;		/* still in quoted block */
-		if (!sep && p > start && *p == '=')
-			sep = p;		/* name and value separator */
-		if (*p == ',')
-			stop = p;		/* terminate the option item */
-		else if (*(p + 1) == '\0')
-			stop = p + 1;		/* end of optstr */
-		if (!start || !stop)
-			continue;
-		if (stop <= start)
-			goto error;
-
-		if (name)
-			*name = start;
-		if (namesz)
-			*namesz = sep ? sep - start : stop - start;
-		*optstr = *stop ? stop + 1 : stop;
-
-		if (sep) {
-			if (value)
-				*value = sep + 1;
-			if (valsz)
-				*valsz = stop - sep - 1;
-		}
-		return 0;
-	}
-
-	return 1;				/* end of optstr */
-
-error:
-	DBG(OPTIONS, ul_debug("parse error: \"%s\"", optstr0));
-	return -EINVAL;
-}
-
 /*
  * Locates the first option that matches @name. The @end is set to the
  * char behind the option (it means ',' or \0).
@@ -129,9 +58,11 @@ static int mnt_optstr_locate_option(char *optstr, const char *name,
 	assert(name);
 
 	namesz = strlen(name);
+	if (!namesz)
+		return 1;
 
 	do {
-		rc = mnt_optstr_parse_next(&optstr, &n, &nsz,
+		rc = ul_optstr_next(&optstr, &n, &nsz,
 					&ol->value, &ol->valsz);
 		if (rc)
 			break;
@@ -165,50 +96,35 @@ int mnt_optstr_next_option(char **optstr, char **name, size_t *namesz,
 {
 	if (!optstr || !*optstr)
 		return -EINVAL;
-	return mnt_optstr_parse_next(optstr, name, namesz, value, valuesz);
+
+	return ul_optstr_next(optstr, name, namesz, value, valuesz);
 }
 
-static int __mnt_optstr_append_option(char **optstr,
-			const char *name, size_t nsz,
-			const char *value, size_t vsz)
+int mnt_buffer_append_option(struct ul_buffer *buf,
+			const char *name, size_t namesz,
+			const char *val, size_t valsz,
+			int quoted)
 {
-	char *p;
-	size_t sz, osz;
+	int rc = 0;
 
-	assert(name);
-	assert(*name);
-	assert(nsz);
-	assert(optstr);
-
-	osz = *optstr ? strlen(*optstr) : 0;
-
-	sz = osz + nsz + 1;		/* 1: '\0' */
-	if (osz)
-		sz++;			/* ',' options separator */
-	if (value)
-		sz += vsz + 1;		/* 1: '=' */
-
-	p = realloc(*optstr, sz);
-	if (!p)
-		return -ENOMEM;
-	*optstr = p;
-
-	if (osz) {
-		p += osz;
-		*p++ = ',';
+	if (!ul_buffer_is_empty(buf))
+		rc = ul_buffer_append_data(buf, ",", 1);
+	if (!rc)
+		rc = ul_buffer_append_data(buf, name, namesz);
+	if (val && !rc) {
+		/* we need to append '=' is value is empty string, see
+		 * 727c689908c5e68c92aa1dd65e0d3bdb6d91c1e5 */
+		rc = ul_buffer_append_data(buf, "=", 1);
+		if (!rc && valsz) {
+			if (quoted)
+				rc = ul_buffer_append_data(buf, "\"", 1);
+			if (!rc)
+				rc = ul_buffer_append_data(buf, val, valsz);
+			if (quoted)
+				rc = ul_buffer_append_data(buf, "\"", 1);
+		}
 	}
-
-	memcpy(p, name, nsz);
-	p += nsz;
-
-	if (value) {
-		*p++ = '=';
-		memcpy(p, value, vsz);
-		p += vsz;
-	}
-	*p = '\0';
-
-	return 0;
+	return rc;
 }
 
 /**
@@ -222,7 +138,9 @@ static int __mnt_optstr_append_option(char **optstr,
  */
 int mnt_optstr_append_option(char **optstr, const char *name, const char *value)
 {
-	size_t vsz, nsz;
+	struct ul_buffer buf = UL_INIT_BUFFER;
+	int rc;
+	size_t nsz, vsz, osz;
 
 	if (!optstr)
 		return -EINVAL;
@@ -230,11 +148,20 @@ int mnt_optstr_append_option(char **optstr, const char *name, const char *value)
 		return 0;
 
 	nsz = strlen(name);
+	osz = *optstr ? strlen(*optstr) : 0;
 	vsz = value ? strlen(value) : 0;
 
-	return __mnt_optstr_append_option(optstr, name, nsz, value, vsz);
-}
+	ul_buffer_refer_string(&buf, *optstr);
+	ul_buffer_set_chunksize(&buf, osz + nsz + vsz + 3);	/* to call realloc() only once */
 
+	rc = mnt_buffer_append_option(&buf, name, nsz, value, vsz, 0);
+	if (!rc)
+		*optstr = ul_buffer_get_data(&buf, NULL, NULL);
+	else if (osz == 0)
+		ul_buffer_free_data(&buf);
+
+	return rc;
+}
 /**
  * mnt_optstr_prepend_option:
  * @optstr: option string or NULL, returns a reallocated string
@@ -246,30 +173,34 @@ int mnt_optstr_append_option(char **optstr, const char *name, const char *value)
  */
 int mnt_optstr_prepend_option(char **optstr, const char *name, const char *value)
 {
-	int rc = 0;
-	char *tmp;
+	struct ul_buffer buf = UL_INIT_BUFFER;
+	size_t nsz, vsz, osz;
+	int rc;
 
 	if (!optstr)
 		return -EINVAL;
 	if (!name || !*name)
 		return 0;
 
-	tmp = *optstr;
-	*optstr = NULL;
+	nsz = strlen(name);
+	osz = *optstr ? strlen(*optstr) : 0;
+	vsz = value ? strlen(value) : 0;
 
-	rc = mnt_optstr_append_option(optstr, name, value);
-	if (!rc && tmp && *tmp)
-		rc = mnt_optstr_append_option(optstr, tmp, NULL);
-	if (!rc) {
-		free(tmp);
-		return 0;
+	ul_buffer_set_chunksize(&buf, osz + nsz + vsz + 3);   /* to call realloc() only once */
+
+	rc = mnt_buffer_append_option(&buf, name, nsz, value, vsz, 0);
+	if (*optstr && !rc) {
+		rc = ul_buffer_append_data(&buf, ",", 1);
+		if (!rc)
+			rc = ul_buffer_append_data(&buf, *optstr, osz);
+		free(*optstr);
 	}
 
-	free(*optstr);
-	*optstr = tmp;
+	if (!rc)
+		*optstr = ul_buffer_get_data(&buf, NULL, NULL);
+	else
+		ul_buffer_free_data(&buf);
 
-	DBG(OPTIONS, ul_debug("failed to prepend '%s[=%s]' to '%s'",
-				name, value, *optstr));
 	return rc;
 }
 
@@ -286,14 +217,12 @@ int mnt_optstr_prepend_option(char **optstr, const char *name, const char *value
 int mnt_optstr_get_option(const char *optstr, const char *name,
 			  char **value, size_t *valsz)
 {
-	struct libmnt_optloc ol;
+	struct libmnt_optloc ol = MNT_INIT_OPTLOC;
 	int rc;
 
 	if (!optstr || !name)
 		return -EINVAL;
 
-	mnt_init_optloc(&ol);
-
 	rc = mnt_optstr_locate_option((char *) optstr, name, &ol);
 	if (!rc) {
 		if (value)
@@ -324,9 +253,7 @@ int mnt_optstr_deduplicate_option(char **optstr, const char *name)
 
 	opt = *optstr;
 	do {
-		struct libmnt_optloc ol;
-
-		mnt_init_optloc(&ol);
+		struct libmnt_optloc ol = MNT_INIT_OPTLOC;
 
 		rc = mnt_optstr_locate_option(opt, name, &ol);
 		if (!rc) {
@@ -345,7 +272,9 @@ int mnt_optstr_deduplicate_option(char **optstr, const char *name)
 			end = ol.end;
 			opt = end && *end ? end + 1 : NULL;
 		}
-	} while (rc == 0 && opt && *opt);
+		if (opt == NULL)
+			break;
+	} while (rc == 0 && *opt);
 
 	return rc < 0 ? rc : begin ? 0 : 1;
 }
@@ -381,7 +310,7 @@ insert_value(char **str, char *pos, const char *substr, char **next)
 	size_t strsz = strlen(*str);
 	size_t possz = strlen(pos);
 	size_t posoff;
-	char *p = NULL;
+	char *p;
 	int sep;
 
 	/* is it necessary to prepend '=' before the substring ? */
@@ -431,15 +360,13 @@ insert_value(char **str, char *pos, const char *substr, char **next)
  */
 int mnt_optstr_set_option(char **optstr, const char *name, const char *value)
 {
-	struct libmnt_optloc ol;
+	struct libmnt_optloc ol = MNT_INIT_OPTLOC;
 	char *nameend;
 	int rc = 1;
 
 	if (!optstr || !name)
 		return -EINVAL;
 
-	mnt_init_optloc(&ol);
-
 	if (*optstr)
 		rc = mnt_optstr_locate_option(*optstr, name, &ol);
 	if (rc < 0)
@@ -478,14 +405,12 @@ int mnt_optstr_set_option(char **optstr, const char *name, const char *value)
  */
 int mnt_optstr_remove_option(char **optstr, const char *name)
 {
-	struct libmnt_optloc ol;
+	struct libmnt_optloc ol = MNT_INIT_OPTLOC;
 	int rc;
 
 	if (!optstr || !name)
 		return -EINVAL;
 
-	mnt_init_optloc(&ol);
-
 	rc = mnt_optstr_locate_option(*optstr, name, &ol);
 	if (rc != 0)
 		return rc;
@@ -518,9 +443,13 @@ int mnt_optstr_remove_option(char **optstr, const char *name)
 int mnt_split_optstr(const char *optstr, char **user, char **vfs,
 		     char **fs, int ignore_user, int ignore_vfs)
 {
+	int rc = 0;
 	char *name, *val, *str = (char *) optstr;
-	size_t namesz, valsz;
+	size_t namesz, valsz, chunsz;
 	struct libmnt_optmap const *maps[2];
+	struct ul_buffer xvfs = UL_INIT_BUFFER,
+			 xfs = UL_INIT_BUFFER,
+			 xuser = UL_INIT_BUFFER;
 
 	if (!optstr)
 		return -EINVAL;
@@ -528,15 +457,10 @@ int mnt_split_optstr(const char *optstr, char **user, char **vfs,
 	maps[0] = mnt_get_builtin_optmap(MNT_LINUX_MAP);
 	maps[1] = mnt_get_builtin_optmap(MNT_USERSPACE_MAP);
 
-	if (vfs)
-		*vfs = NULL;
-	if (fs)
-		*fs = NULL;
-	if (user)
-		*user = NULL;
+	chunsz = strlen(optstr) / 2;
 
 	while (!mnt_optstr_next_option(&str, &name, &namesz, &val, &valsz)) {
-		int rc = 0;
+		struct ul_buffer *buf = NULL;
 		const struct libmnt_optmap *ent = NULL;
 		const struct libmnt_optmap *m =
 			 mnt_optmap_get_entry(maps, 2, name, namesz, &ent);
@@ -551,34 +475,40 @@ int mnt_split_optstr(const char *optstr, char **user, char **vfs,
 		if (ent && m && m == maps[0] && vfs) {
 			if (ignore_vfs && (ent->mask & ignore_vfs))
 				continue;
-			rc = __mnt_optstr_append_option(vfs, name, namesz,
-								val, valsz);
+			if (vfs)
+				buf = &xvfs;
 		} else if (ent && m && m == maps[1] && user) {
 			if (ignore_user && (ent->mask & ignore_user))
 				continue;
-			rc = __mnt_optstr_append_option(user, name, namesz,
-								val, valsz);
-		} else if (!m && fs)
-			rc = __mnt_optstr_append_option(fs, name, namesz,
-								val, valsz);
-		if (rc) {
-			if (vfs) {
-				free(*vfs);
-				*vfs = NULL;
-			}
-			if (fs) {
-				free(*fs);
-				*fs = NULL;
-			}
-			if (user) {
-				free(*user);
-				*user = NULL;
-			}
-			return rc;
+			if (user)
+				buf = &xuser;
+		} else if (!m && fs) {
+			if (fs)
+				buf = &xfs;
 		}
+
+		if (buf) {
+			if (ul_buffer_is_empty(buf))
+				ul_buffer_set_chunksize(buf, chunsz);
+			rc = mnt_buffer_append_option(buf, name, namesz, val, valsz, 0);
+		}
+		if (rc)
+			break;
 	}
 
-	return 0;
+	if (vfs)
+		*vfs  = rc ? NULL : ul_buffer_get_data(&xvfs, NULL, NULL);
+	if (fs)
+		*fs   = rc ? NULL : ul_buffer_get_data(&xfs, NULL, NULL);
+	if (user)
+		*user = rc ? NULL : ul_buffer_get_data(&xuser, NULL, NULL);
+	if (rc) {
+		ul_buffer_free_data(&xvfs);
+		ul_buffer_free_data(&xfs);
+		ul_buffer_free_data(&xuser);
+	}
+
+	return rc;
 }
 
 /**
@@ -603,17 +533,19 @@ int mnt_optstr_get_options(const char *optstr, char **subset,
 			    const struct libmnt_optmap *map, int ignore)
 {
 	struct libmnt_optmap const *maps[1];
+	struct ul_buffer buf = UL_INIT_BUFFER;
 	char *name, *val, *str = (char *) optstr;
 	size_t namesz, valsz;
+	int rc = 0;
 
 	if (!optstr || !subset)
 		return -EINVAL;
 
 	maps[0] = map;
-	*subset = NULL;
 
-	while(!mnt_optstr_next_option(&str, &name, &namesz, &val, &valsz)) {
-		int rc = 0;
+	ul_buffer_set_chunksize(&buf, strlen(optstr)/2);
+
+	while (!mnt_optstr_next_option(&str, &name, &namesz, &val, &valsz)) {
 		const struct libmnt_optmap *ent;
 
 		mnt_optmap_get_entry(maps, 1, name, namesz, &ent);
@@ -628,14 +560,15 @@ int mnt_optstr_get_options(const char *optstr, char **subset,
 		if (valsz && mnt_optmap_entry_novalue(ent))
 			continue;
 
-		rc = __mnt_optstr_append_option(subset, name, namesz, val, valsz);
-		if (rc) {
-			free(*subset);
-			return rc;
-		}
+		rc = mnt_buffer_append_option(&buf, name, namesz, val, valsz, 0);
+		if (rc)
+			break;
 	}
 
-	return 0;
+	*subset  = rc ? NULL : ul_buffer_get_data(&buf, NULL, NULL);
+	if (rc)
+		ul_buffer_free_data(&buf);
+	return rc;
 }
 
 
@@ -724,20 +657,22 @@ int mnt_optstr_get_flags(const char *optstr, unsigned long *flags,
  *	MS_NOATIME and "foo,bar,noexec"   --returns->  "foo,bar,noatime"
  *
  * Returns: 0 on success or negative number in case of error.
+ *
+ * Deprecated: since v2.39.
  */
 int mnt_optstr_apply_flags(char **optstr, unsigned long flags,
 				const struct libmnt_optmap *map)
 {
 	struct libmnt_optmap const *maps[1];
 	char *name, *next, *val;
-	size_t namesz = 0, valsz = 0;
+	size_t namesz = 0, valsz = 0, multi = 0;
 	unsigned long fl;
 	int rc = 0;
 
 	if (!optstr || !map)
 		return -EINVAL;
 
-	DBG(CXT, ul_debug("applying 0x%08lu flags to '%s'", flags, *optstr));
+	DBG(CXT, ul_debug("applying 0x%08lx flags to '%s'", flags, *optstr));
 
 	maps[0] = map;
 	next = *optstr;
@@ -799,17 +734,32 @@ int mnt_optstr_apply_flags(char **optstr, unsigned long flags,
 					if (rc)
 						goto err;
 				}
-				if (!(ent->mask & MNT_INVERT))
-					fl &= ~ent->id;
+				if (!(ent->mask & MNT_INVERT)) {
+					/* allow options with prefix (X-mount.foo,X-mount.bar) more than once */
+					if (ent->mask & MNT_PREFIX)
+						multi |= ent->id;
+					else
+						fl &= ~ent->id;
+					if (ent->id & MS_REC)
+						fl |= MS_REC;
+				}
 			}
 		}
 	}
 
-	/* add missing options */
-	if (fl) {
+	/* remove from flags options which are allowed more than once */
+	fl &= ~multi;
+
+	/* add missing options (but ignore fl if contains MS_REC only) */
+	if (fl && fl != MS_REC) {
+
 		const struct libmnt_optmap *ent;
+		struct ul_buffer buf = UL_INIT_BUFFER;
+		size_t sz;
 		char *p;
 
+		ul_buffer_refer_string(&buf, *optstr);
+
 		for (ent = map; ent && ent->name; ent++) {
 			if ((ent->mask & MNT_INVERT)
 			    || ent->id == 0
@@ -823,17 +773,20 @@ int mnt_optstr_apply_flags(char **optstr, unsigned long flags,
 					p--;			/* name[=] */
 				else
 					continue;		/* name= */
-
-				p = strndup(ent->name, p - ent->name);
-				if (!p) {
-					rc = -ENOMEM;
-					goto err;
-				}
-				mnt_optstr_append_option(optstr, p, NULL);
-				free(p);
+				sz = p - ent->name;
 			} else
-				mnt_optstr_append_option(optstr, ent->name, NULL);
+				sz = strlen(ent->name);
+
+			rc = mnt_buffer_append_option(&buf, ent->name, sz, NULL, 0, 0);
+			if (rc)
+				break;
 		}
+
+		if (rc) {
+			ul_buffer_free_data(&buf);
+			goto err;
+		} else
+			*optstr = ul_buffer_get_data(&buf, NULL, NULL);
 	}
 
 	DBG(CXT, ul_debug("new optstr '%s'", *optstr));
@@ -843,244 +796,120 @@ err:
 	return rc;
 }
 
-/*
- * @optstr: string with comma separated list of options
- * @value: pointer to the begin of the context value
- * @valsz: size of the value
- * @next: returns pointer to the next option (optional argument)
+/**
+ * mnt_match_options:
+ * @optstr: options string
+ * @pattern: comma delimited list of options
  *
- * Translates SELinux context from human to raw format. The function does not
- * modify @optstr and returns zero if libmount is compiled without SELinux
- * support.
+ * The "no" could be used for individual items in the @options list. The "no"
+ * prefix does not have a global meaning.
  *
- * Returns: 0 on success, a negative number in case of error.
+ * Unlike fs type matching, nonetdev,user and nonetdev,nouser have
+ * DIFFERENT meanings; each option is matched explicitly as specified.
+ *
+ * The "no" prefix interpretation could be disabled by the "+" prefix, for example
+ * "+noauto" matches if @optstr literally contains the "noauto" string.
+ *
+ * The alone "no" is error and all matching ends with False.
+ *
+ * "xxx,yyy,zzz" : "nozzz"	-> False
+ *
+ * "xxx,yyy,zzz" : "xxx,noeee"	-> True
+ *
+ * "bar,zzz"     : "nofoo"      -> True		(does not contain "foo")
+ *
+ * "nofoo,bar"   : "nofoo"      -> True		(does not contain "foo")
+ *
+ * "nofoo,bar"   : "+nofoo"     -> True		(contains "nofoo")
+ *
+ * "bar,zzz"     : "+nofoo"     -> False	(does not contain "nofoo")
+ *
+ * "bar,zzz"     : "" or  "+"   -> True		(empty pattern is matching)
+ *
+ * ""            : ""           -> True
+ *
+ * ""            : "foo"        -> False
+ *
+ * ""            : "nofoo"      -> True
+ *
+ * ""            : "no,foo"     -> False	(alone "no" is error)
+ *
+ * "no"          : "+no"        -> True		("no" is an option due to "+")
+ *
+ * Returns: 1 if pattern is matching, else 0. This function also returns 0
+ *          if @pattern is NULL and @optstr is non-NULL.
  */
-#ifndef HAVE_LIBSELINUX
-int mnt_optstr_fix_secontext(char **optstr __attribute__ ((__unused__)),
-			     char *value   __attribute__ ((__unused__)),
-			     size_t valsz  __attribute__ ((__unused__)),
-			     char **next   __attribute__ ((__unused__)))
+int mnt_match_options(const char *optstr, const char *pattern)
 {
-	return 0;
-}
-#else
-int mnt_optstr_fix_secontext(char **optstr,
-			     char *value,
-			     size_t valsz,
-			     char **next)
-{
-	int rc = 0;
+	char *name, *pat = (char *) pattern;
+	char *buf = NULL, *patval;
+	size_t namesz = 0, patvalsz = 0;
+	int match = 1;
 
-	security_context_t raw = NULL;
-	char *p, *val, *begin, *end;
-	size_t sz;
+	if (!pattern && !optstr)
+		return 1;
+	if (pattern && optstr && !*pattern && !*optstr)
+		return 1;
+	if (!pattern)
+		return 0;
 
-	if (!optstr || !*optstr || !value || !valsz)
-		return -EINVAL;
+	/* walk on pattern string
+	 */
+	while (match && !mnt_optstr_next_option(&pat, &name, &namesz,
+						&patval, &patvalsz)) {
+		char *val;
+		size_t sz = 0;
+		int no = 0, rc;
 
-	DBG(CXT, ul_debug("fixing SELinux context"));
+		if (*name == '+')
+			name++, namesz--;
+		else if ((no = (startswith(name, "no") != NULL))) {
+			name += 2, namesz -= 2;
+			if (!*name || *name == ',') {
+				match = 0;
+				break;	/* alone "no" keyword is error */
+			}
+		}
 
-	begin = value;
-	end = value + valsz;
+		if (optstr && *optstr && *name) {
+			if (!buf) {
+				buf = malloc(strlen(pattern) + 1);
+				if (!buf)
+					return 0;
+			}
 
-	/* the selinux contexts are quoted */
-	if (*value == '"') {
-		if (valsz <= 2 || *(value + valsz - 1) != '"')
-			return -EINVAL;		/* improperly quoted option string */
-		value++;
-		valsz -= 2;
+			xstrncpy(buf, name, namesz + 1);
+			rc = mnt_optstr_get_option(optstr, buf, &val, &sz);
+
+		} else if (!*name) {
+			rc = 0;		/* empty pattern matches */
+		} else {
+			rc = 1;		/* not found in empty string */
+		}
+
+		/* check also value (if the pattern is "foo=value") */
+		if (rc == 0 && patvalsz > 0 &&
+		    (patvalsz != sz || strncmp(patval, val, sz) != 0))
+			rc = 1;
+
+		switch (rc) {
+		case 0:		/* found */
+			match = no == 0 ? 1 : 0;
+			break;
+		case 1:		/* not found */
+			match = no == 1 ? 1 : 0;
+			break;
+		default:	/* parse error */
+			match = 0;
+			break;
+		}
 	}
 
-	p = strndup(value, valsz);
-	if (!p)
-		return -ENOMEM;
-
-
-	/* translate the context */
-	rc = selinux_trans_to_raw_context((security_context_t) p, &raw);
-
-	DBG(CXT, ul_debug("SELinux context '%s' translated to '%s'",
-			p, rc == -1 ? "FAILED" : (char *) raw));
-
-	free(p);
-	if (rc == -1 ||	!raw)
-		return -EINVAL;
-
-
-	/* create a quoted string from the raw context */
-	sz = strlen((char *) raw);
-	if (!sz)
-		return -EINVAL;
-
-	p = val = malloc(valsz + 3);
-	if (!val)
-		return -ENOMEM;
-
-	*p++ = '"';
-	memcpy(p, raw, sz);
-	p += sz;
-	*p++ = '"';
-	*p = '\0';
-
-	freecon(raw);
-
-	/* set new context */
-	mnt_optstr_remove_option_at(optstr, begin, end);
-	rc = insert_value(optstr, begin, val, next);
-	free(val);
-
-	return rc;
-}
-#endif
-
-static int set_uint_value(char **optstr, unsigned int num,
-			char *begin, char *end, char **next)
-{
-	char buf[40];
-	snprintf(buf, sizeof(buf), "%u", num);
-
-	mnt_optstr_remove_option_at(optstr, begin, end);
-	return insert_value(optstr, begin, buf, next);
-}
-
-/*
- * @optstr: string with a comma separated list of options
- * @value: pointer to the beginning of the uid value
- * @valsz: size of the value
- * @next: returns pointer to the next option (optional argument)
-
- * Translates "username" or "useruid" to the real UID.
- *
- * For example:
- *	if (!mnt_optstr_get_option(optstr, "uid", &val, &valsz))
- *		mnt_optstr_fix_uid(&optstr, val, valsz, NULL);
- *
- * Returns: 0 on success, a negative number in case of error.
- */
-int mnt_optstr_fix_uid(char **optstr, char *value, size_t valsz, char **next)
-{
-	char *end;
-
-	if (!optstr || !*optstr || !value || !valsz)
-		return -EINVAL;
-
-	DBG(CXT, ul_debug("fixing uid"));
-
-	end = value + valsz;
-
-	if (valsz == 7 && !strncmp(value, "useruid", 7) &&
-	    (*(value + 7) == ',' || !*(value + 7)))
-		return set_uint_value(optstr, getuid(), value, end, next);
-
-	else if (!isdigit(*value)) {
-		uid_t id;
-		int rc;
-		char *p = strndup(value, valsz);
-		if (!p)
-			return -ENOMEM;
-		rc = mnt_get_uid(p, &id);
-		free(p);
-
-		if (!rc)
-			return set_uint_value(optstr, id, value, end, next);
-	}
-
-	if (next) {
-		/* no change, let's keep the original value */
-		*next = value + valsz;
-		if (**next == ',')
-			(*next)++;
-	}
-
-	return 0;
-}
-
-/*
- * @optstr: string with a comma separated list of options
- * @value: pointer to the beginning of the uid value
- * @valsz: size of the value
- * @next: returns pointer to the next option (optional argument)
-
- * Translates "groupname" or "usergid" to the real GID.
- *
- * Returns: 0 on success, a negative number in case of error.
- */
-int mnt_optstr_fix_gid(char **optstr, char *value, size_t valsz, char **next)
-{
-	char *end;
-
-	if (!optstr || !*optstr || !value || !valsz)
-		return -EINVAL;
-
-	DBG(CXT, ul_debug("fixing gid"));
-
-	end = value + valsz;
-
-	if (valsz == 7 && !strncmp(value, "usergid", 7) &&
-	    (*(value + 7) == ',' || !*(value + 7)))
-		return set_uint_value(optstr, getgid(), value, end, next);
-
-	else if (!isdigit(*value)) {
-		int rc;
-		gid_t id;
-		char *p = strndup(value, valsz);
-		if (!p)
-			return -ENOMEM;
-		rc = mnt_get_gid(p, &id);
-		free(p);
-
-		if (!rc)
-			return set_uint_value(optstr, id, value, end, next);
-
-	}
-
-	if (next) {
-		/* nothing */
-		*next = value + valsz;
-		if (**next == ',')
-			(*next)++;
-	}
-	return 0;
-}
-
-/*
- * Converts "user" to "user=<username>".
- *
- * Returns: 0 on success, negative number in case of error.
- */
-int mnt_optstr_fix_user(char **optstr)
-{
-	char *username;
-	struct libmnt_optloc ol;
-	int rc = 0;
-
-	DBG(CXT, ul_debug("fixing user"));
-
-	mnt_init_optloc(&ol);
-
-	rc = mnt_optstr_locate_option(*optstr, "user", &ol);
-	if (rc)
-		return rc == 1 ? 0 : rc;	/* 1: user= not found */
-
-	username = mnt_get_username(getuid());
-	if (!username)
-		return -ENOMEM;
-
-	if (!ol.valsz || (ol.value && strncmp(ol.value, username, ol.valsz))) {
-		if (ol.valsz)
-			/* remove old value */
-			mnt_optstr_remove_option_at(optstr, ol.value, ol.end);
-
-		rc = insert_value(optstr, ol.value ? ol.value : ol.end,
-				  username, NULL);
-	}
-
-	free(username);
-	return rc;
+	free(buf);
+	return match;
 }
 
 #ifdef TEST_PROGRAM
-
 static int test_append(struct libmnt_test *ts, int argc, char *argv[])
 {
 	const char *value = NULL, *name;
@@ -1090,6 +919,8 @@ static int test_append(struct libmnt_test *ts, int argc, char *argv[])
 	if (argc < 3)
 		return -EINVAL;
 	optstr = strdup(argv[1]);
+	if (!optstr)
+		err_oom();
 	name = argv[2];
 
 	if (argc == 4)
@@ -1111,6 +942,8 @@ static int test_prepend(struct libmnt_test *ts, int argc, char *argv[])
 	if (argc < 3)
 		return -EINVAL;
 	optstr = strdup(argv[1]);
+	if (!optstr)
+		err_oom();
 	name = argv[2];
 
 	if (argc == 4)
@@ -1132,6 +965,8 @@ static int test_split(struct libmnt_test *ts, int argc, char *argv[])
 		return -EINVAL;
 
 	optstr = strdup(argv[1]);
+	if (!optstr)
+		err_oom();
 
 	rc = mnt_split_optstr(optstr, &user, &vfs, &fs, 0, 0);
 	if (!rc) {
@@ -1157,6 +992,8 @@ static int test_flags(struct libmnt_test *ts, int argc, char *argv[])
 		return -EINVAL;
 
 	optstr = strdup(argv[1]);
+	if (!optstr)
+		err_oom();
 
 	rc = mnt_optstr_get_flags(optstr, &fl, mnt_get_builtin_optmap(MNT_LINUX_MAP));
 	if (rc)
@@ -1192,6 +1029,8 @@ static int test_apply(struct libmnt_test *ts, int argc, char *argv[])
 	}
 
 	optstr = strdup(argv[2]);
+	if (!optstr)
+		err_oom();
 	flags = strtoul(argv[3], NULL, 16);
 
 	printf("flags:  0x%08lx\n", flags);
@@ -1212,6 +1051,8 @@ static int test_set(struct libmnt_test *ts, int argc, char *argv[])
 	if (argc < 3)
 		return -EINVAL;
 	optstr = strdup(argv[1]);
+	if (!optstr)
+		err_oom();
 	name = argv[2];
 
 	if (argc == 4)
@@ -1262,6 +1103,8 @@ static int test_remove(struct libmnt_test *ts, int argc, char *argv[])
 	if (argc < 3)
 		return -EINVAL;
 	optstr = strdup(argv[1]);
+	if (!optstr)
+		err_oom();
 	name = argv[2];
 
 	rc = mnt_optstr_remove_option(&optstr, name);
@@ -1280,6 +1123,8 @@ static int test_dedup(struct libmnt_test *ts, int argc, char *argv[])
 	if (argc < 3)
 		return -EINVAL;
 	optstr = strdup(argv[1]);
+	if (!optstr)
+		err_oom();
 	name = argv[2];
 
 	rc = mnt_optstr_deduplicate_option(&optstr, name);
@@ -1289,39 +1134,19 @@ static int test_dedup(struct libmnt_test *ts, int argc, char *argv[])
 	return rc;
 }
 
-static int test_fix(struct libmnt_test *ts, int argc, char *argv[])
+static int test_match(struct libmnt_test *ts, int argc, char *argv[])
 {
-	char *optstr;
-	int rc = 0;
-	char *name, *val, *next;
-	size_t valsz, namesz;
+	char *optstr, *pattern;
 
-	if (argc < 2)
+	if (argc < 3)
 		return -EINVAL;
 
-	next = optstr = strdup(argv[1]);
-
-	printf("optstr: %s\n", optstr);
-
-	while (!mnt_optstr_next_option(&next, &name, &namesz, &val, &valsz)) {
-
-		if (!strncmp(name, "uid", 3))
-			rc = mnt_optstr_fix_uid(&optstr, val, valsz, &next);
-		else if (!strncmp(name, "gid", 3))
-			rc = mnt_optstr_fix_gid(&optstr, val, valsz, &next);
-		else if (!strncmp(name, "context", 7))
-			rc = mnt_optstr_fix_secontext(&optstr, val, valsz, &next);
-		if (rc)
-			break;
-	}
-	if (rc)
-		rc = mnt_optstr_fix_user(&optstr);
-
-	printf("fixed:  %s\n", optstr);
-
-	free(optstr);
-	return rc;
-
+	optstr = argv[1];
+	pattern = argv[2];
+	printf("%-6s: \"%s\"\t:\t\"%s\"\n",
+			mnt_match_options(optstr, pattern) == 1 ? "true" : "false",
+			optstr, pattern);
+	return 0;
 }
 
 int main(int argc, char *argv[])
@@ -1333,10 +1158,10 @@ int main(int argc, char *argv[])
 		{ "--get",    test_get,    "<optstr> <name>            search name in optstr" },
 		{ "--remove", test_remove, "<optstr> <name>            remove name in optstr" },
 		{ "--dedup",  test_dedup,  "<optstr> <name>            deduplicate name in optstr" },
+		{ "--match",  test_match,  "<optstr> <pattern>         compare optstr with pattern" },
 		{ "--split",  test_split,  "<optstr>                   split into FS, VFS and userspace" },
 		{ "--flags",  test_flags,  "<optstr>                   convert options to MS_* flags" },
 		{ "--apply",  test_apply,  "--{linux,user} <optstr> <mask>    apply mask to optstr" },
-		{ "--fix",    test_fix,    "<optstr>                   fix uid=, gid=, user, and context=" },
 
 		{ NULL }
 	};
